/* #[info:			*/
/************************************************************************
*									*
*									*
*		   _	       _     _					*
*		  / \\	/'\\  / \\  < \, \\/\\/\\  _-_			*
*		 || || || || || ||  /-|| || || || || \\			*
*		 || || || || || || (( || || || || ||/			*
*		 \\_-| \\,/  \\_-|  \/\\ \\ \\ \\ \\,/			*
*		  /  \	      /	 \					*
*		 '----`	     '----`					*
*									*
*		  Copyright (c) 1993  Jan van der Steen			*
*	       Centre for Mathematics and Computer Science		*
*		       Amsterdam, the Netherlands			*
*									*
*-----------------------------------------------------------------------*
*   You are not allowed to use this software in commercial software.	*
*   You are allowed to use this software in free software under the	*
*   condition that you maintain this integral message in the source.	*
*-----------------------------------------------------------------------*
* File	  : gogame.c							*
* Purpose : The Go game data structure					*
* Version : 1.31							*
* Modified: 22 Mar 1995 09:43:44					*
* Author  : Jan van der Steen (jansteen@cwi.nl)				*
************************************************************************/
/* #]info:			*/ 
/* #[include:			*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __STRICT_ANSI__
char * strdup           (const char *);
int    strncasecmp      (const char *s1, const char *s2, size_t n);
int    isascii          (int c);
#endif

#include "tools.h"
#include "gogame.h"
#include "gorules.h"
#ifdef memcpy
#undef memcpy
#endif

/* #]include:			*/ 
/* #[memory pools:		*/

#include "mempool.h"
MP *mpool = NULL;	/* move  pool  */
MP *cpool = NULL;	/* chain pool */
MP *tpool = NULL;	/* text  pool */

/* #]memory pools:		*/ 
/* #[prototype: 		*/

#ifdef __STDC__
#	define	PROTO(s) s
#else
#	define	PROTO(s) ()
#endif

/*
 * Memory managers
 */
static CHAIN *	chain_alloc	PROTO((void));
static TEXT *	text_alloc	PROTO((void));
static GOBAN *	goban_alloc	PROTO((int size));

static	void	_ch_free	PROTO((CHAIN  *c, int f));
static	void	_mv_free	PROTO((GOMOVE *m));
static	void	_inf_free	PROTO((GOGAME *g));
static	void	_txt_free	PROTO((TEXT   *t));

/*
 * Some events need a move struct to be stored
 */
static GOMOVE *	need_move	PROTO((GOGAME *g, char *why));

#undef PROTO

/* #]prototype: 		*/ 

/* #[game_alloc:		*/

GOGAME *
game_alloc(g)
/*
 * Allocate a (sub)game
 */
GOGAME *g;
{
    GOGAME *h = (GOGAME *) calloc(1, sizeof(GOGAME));

    if (h == NULL) {
	message("Out of memory in game_alloc()\n");
	return NULL;
    }

    if (g == NULL) {
	/*
	 * Allocate a Go board
	 */
	if ((h->gm_goban = goban_new(h, MAXSIZE)) == NULL) {
	    h->gm_sz = 0;
	    return NULL;
	}
	h->gm_sz = MAXSIZE;
    }
    else {
	/*
	 * Use the parents Go board
	 */
	h->gm_goban = g->gm_goban;
	h->gm_sz  = g->gm_sz;
	memcpy(h->gm_vw, g->gm_vw, sizeof(h->gm_vw));
    }

    return h;
}

/* #]game_alloc:		*/ 
/* #[game_free: 		*/

void
game_free(g)
/*
 * Free the resources associated with a game
 */
GOGAME *g;
{
    if (g) {
	if (g->gm_next) game_free(g->gm_next);
	if (g->gm_move) _mv_free(g->gm_move);
	/*
	 * Only free goban of parent, children use parents board
	 */
	if (!g->gm_parent && g->gm_goban) {
	    if (g->gm_goban->gb_node) free(g->gm_goban->gb_node);
	    free(g->gm_goban);
	}
	_inf_free(g);
	free(g);
    }
}

/* #[_ch_free:		*/

static void
_ch_free(c, f)
CHAIN *c;
int    f;
{
    if (c) {
	if (c->ch_next         ) _ch_free(c->ch_next, f);
	if (f && c->ch_newstone)  mp_free(mpool, c->ch_newstone);
	mp_free(cpool, c);
    }
}

/* #]_ch_free:		*/ 
/* #[_txt_free:		*/

static void
_txt_free(t)
TEXT *t;
{
    if (t) {
	if (t->cc_next) _txt_free(t->cc_next);
	if (t->cc_name)      free(t->cc_name);
	if (t->cc_text)      free(t->cc_text);
	mp_free(tpool, t);
    }
}

/* #]_txt_free:		*/ 
/* #[_mv_free:		*/

static void
_mv_free(m)
GOMOVE *m;
{
    if (m) {
	if (m->mv_next     )  _mv_free(m->mv_next);
	if (m->mv_addstone )  _ch_free(m->mv_addstone, 1);
	if (m->mv_captured )  _ch_free(m->mv_captured, 0);
	if (m->mv_delstone )  _ch_free(m->mv_delstone, 0);
	if (m->mv_letter   )  _ch_free(m->mv_letter  , 0);
	if (m->mv_mark     )  _ch_free(m->mv_mark    , 0);
	if (m->mv_text     ) _txt_free(m->mv_text);
	if (m->mv_variation) game_free(m->mv_variation);
	if (m->mv_btime    )      free(m->mv_btime);
	if (m->mv_wtime    )      free(m->mv_wtime);
	mp_free(mpool, m);
    }
}

/* #]_mv_free:		*/ 
/* #[_inf_free:		*/

static void
_inf_free(g)
/*
 * Free game information
 */
GOGAME *g;
{
    if (g) {
	if (g->gm_gn) free(g->gm_gn);
	if (g->gm_ev   ) free(g->gm_ev);
	if (g->gm_us    ) free(g->gm_us);
	if (g->gm_gamenr  ) free(g->gm_gamenr);
	if (g->gm_n    ) free(g->gm_n);
	if (g->gm_pb   ) free(g->gm_pb);
	if (g->gm_br   ) free(g->gm_br);
	if (g->gm_bl   ) free(g->gm_bl);
	if (g->gm_pw   ) free(g->gm_pw);
	if (g->gm_wr   ) free(g->gm_wr);
	if (g->gm_wl   ) free(g->gm_wl);
	if (g->gm_km    ) free(g->gm_km);
	if (g->gm_dt    ) free(g->gm_dt);
	if (g->gm_pc   ) free(g->gm_pc);
	if (g->gm_re  ) free(g->gm_re);
    }
}

/* #]_inf_free:		*/ 

/* #]game_free: 		*/ 
/* #[goban_alloc:		*/

static GOBAN *
goban_alloc (size)
/*
 * Allocate a new Go board
 */
int size;
{
    GOBAN * goban = (GOBAN *) calloc(1, sizeof(GOBAN));

    if (!goban) {
	message("Out of memory in goban_alloc()\n");
	return NULL;
    }
    if (size <= 0) {
	message("Board size out of range in goban_alloc(): %d\n", size);
	return NULL;
    }
    goban->gb_node = (GONODE *) calloc(size*size, sizeof(GONODE));
    if (!goban->gb_node) {
	message("Out of memory in goban_alloc()\n");
	free(goban);
	return NULL;
    }

    return goban;
}

/* #]goban_alloc:		*/ 
/* #[goban_new:			*/

GOBAN *
goban_new(g, s)
/*
 * Allocate and construct a new Go board
 */
GOGAME *g;
int	s;		/* Requested size */
{
    PLACE  p;

    if ((g->gm_goban = goban_alloc(s)) == NULL) return NULL;
    for (p = 0; p < s*s; p++) {
	NODE(g,p)->gn_n = NODE(g,p-s);
	NODE(g,p)->gn_e = NODE(g,p+1);
	NODE(g,p)->gn_s = NODE(g,p+s);
	NODE(g,p)->gn_w = NODE(g,p-1);
    }
    for (p = 0	    ; p < s  ; p++ ) NODE(g,p)->gn_n = NULL;
    for (p = s-1    ; p < s*s; p+=s) NODE(g,p)->gn_e = NULL;
    for (p = s*(s-1); p < s*s; p++ ) NODE(g,p)->gn_s = NULL;
    for (p = 0	    ; p < s*s; p+=s) NODE(g,p)->gn_w = NULL;

    return g->gm_goban;
}

/* #]goban_new:			*/ 
/* #[board_alloc:		*/

int
board_alloc(g)
/*
 * Allocate slots to maintain the board status while printing
 */
GOGAME *g;
{
    GOGAME *parent = g->gm_parent;
    int	    n      = g->gm_sz*g->gm_sz;

    g->gm_status  = (GOMOVE *) calloc(n, sizeof(GOMOVE));
    g->gm_board   = (GOMOVE *) calloc(n, sizeof(GOMOVE));
    g->gm_initial = (GOMOVE *) calloc(n, sizeof(GOMOVE));

    if (g->gm_status  == NULL ||
	g->gm_board   == NULL ||
	g->gm_initial == NULL) {
	message("Out of memory in board_alloc()\n");
	return 0;
    }

    if (parent && parent->gm_status) {
	memcpy(g->gm_status , parent->gm_status, n*sizeof(GOMOVE));
	memcpy(g->gm_board  , parent->gm_status, n*sizeof(GOMOVE));
	memcpy(g->gm_initial, parent->gm_status, n*sizeof(GOMOVE));
    }
    return 1;
}

/* #]board_alloc:		*/ 
/* #[board_free:		*/

void
board_free(g)
GOGAME *g;
{
    free(g->gm_status ); g->gm_status  = NULL;
    free(g->gm_board  ); g->gm_board   = NULL;
    free(g->gm_initial); g->gm_initial = NULL;
}

/* #]board_free:		*/ 
/* #[move_alloc:		*/

GOMOVE *
move_alloc()
/*
 * Allocate a new move
 */
{
    GOMOVE *m;

    if (!mpool) mpool = mp_init(361, 100, sizeof(GOMOVE), 0);
    if ((m = (GOMOVE *) mp_calloc(mpool)) == NULL) {
	message("Out of memory in move_alloc()\n");
	return NULL;
    }
    m->mv_place = MV_HEADER;
    return m;
}

/* #]move_alloc:		*/ 
/* #[move_new:			*/

GOMOVE *
move_new(g, parent)
GOGAME *g;
GOMOVE *parent;
{
    GOMOVE *m = move_alloc();

    if (!m) return NULL;
    if (!g->gm_move) g->gm_move = m;
    if (parent) {
	parent->mv_next = m;
	m->mv_prev = parent;
    }
    return m;
}

/* #]move_new:			*/ 
/* #[move_free:			*/

GOMOVE *
move_free(m)
/*
 * Remove a move and free it's associated resources
 */
GOMOVE *m;
{
    GOMOVE *prev = m->mv_prev;
    CHAIN  *c, *cn;

    /*
     * Establish the bypass
     */
    if (m->mv_prev) m->mv_prev->mv_next = m->mv_next;
    if (m->mv_next) m->mv_next->mv_prev = m->mv_prev;

    /*
     * Free the resources
     */
    for (c = m->mv_captured; c; c = cn) {
	cn = c->ch_next;
	mp_free(cpool, c);
    }
    for (c = m->mv_delstone; c; c = cn) {
	cn = c->ch_next;
	mp_free(cpool, c);
    }
    for (c = m->mv_addstone; c; c = cn) {
	cn = c->ch_next;
	mp_free(mpool, c->ch_newstone);
	mp_free(cpool, c);
    }
    mp_free(mpool, m);

    return prev;
}

/* #]move_free:			*/ 
/* #[text_alloc:		*/

static TEXT *
text_alloc()
/*
 * Allocate a text list
 */
{
    TEXT *t;

    if (!tpool) tpool = mp_init(100, 100, sizeof(TEXT), 0);
    if ((t = (TEXT *) mp_calloc(tpool)) == NULL) {
	message("Out of memory in text_alloc()\n");
	return NULL;
    }
    return t;
}

/* #]text_alloc:		*/ 
/* #[chain_alloc:		*/

static CHAIN *
chain_alloc()
/*
 * Allocate a new chain
 */
{
    CHAIN *c;

    if (!cpool) cpool = mp_init(361, 100, sizeof(CHAIN), 0);
    if ((c = (CHAIN *) mp_calloc(cpool)) == NULL) {
	message("Out of memory in chain_alloc()\n");
	return NULL;
    }
    return c;
}

/* #]chain_alloc:		*/ 

/* #[text_store:		*/

void
text_store(g, line)
/*
 * We append the text to the text list.
 */
GOGAME *g;
char   *line;
{
    GOMOVE *m;
    TEXT *text, *textlast;

    if (!line || *line == 0) return;
    if ((m = need_move(g, "text")) == NULL) return;
    if ((text = text_alloc()) == NULL) return;

    textlast = m->mv_textlast;
    if (m->mv_text == NULL)
	m->mv_text = text;
    else
	m->mv_textlast->cc_next = text;

    m->mv_textlast = text;
    text->cc_name  = NULL;
    text->cc_spec  = strlen(line);
    text->cc_text  = strdup(line);
    text->cc_prev  = textlast;
}

/* #]text_store:		*/ 
/* #[kibitz_store:		*/

void
kibitz_store(g, line)
/*
 * The way we deal with incomming kibitz text is very IGS specific.
 * Normally we would just append the text to the text buffer.
 * But, since in case of IGS it either contains "say" or "kibitz" messages
 * we have introduced a special format to store these typical cases.
 */
GOGAME *g;
char   *line;
{
    GOMOVE *m = need_move(g, "kibitz");
    TEXT   *text, *textlast;
    char   *name;
    char   *rank;
    char   *mesg;
    char   *space;

    /*
     * The kibitz messages passed here have the format:
     *
     *	<username> <rank><optional [space|*]>: <text>
     *
     *	jansteen 3d: My ~connection~ is terrible...	(old IGS)
     *	jansteen 3d : My ~connection~ is terrible...	(rank)
     *	jansteen 3d*: My ~connection~ is terrible...	(rating)
     */
    if (!m || (text = text_alloc()) == NULL) return;

    textlast = m->mv_textlast;
    if (m->mv_text == NULL)
	m->mv_text = text;
    else
	m->mv_textlast->cc_next = text;
    m->mv_textlast = text;

    /*
     * Let's avoid tests in other parts of the program
     */
    text->cc_name = "";
    text->cc_spec =  0;
    text->cc_text = "";

    name  = line;
    rank  = strchr(name, ' '); if (rank == NULL) goto error;
    mesg  = strchr(rank, ':'); if (mesg == NULL) goto error;
    *rank = 0; rank += 1;
    *mesg = 0; mesg += 2;
    space = rank+strlen(rank)-1; if (*space == ' ') *space = 0;
    text->cc_name = strdup(name);
    text->cc_spec = str2rank    (rank);
    text->cc_text = strdup(mesg);
    text->cc_prev = textlast;
    return;
error:
    message("Bad formatted kibitz: \"%s\" (ignored)\n", line);
}

/* #]kibitz_store:		*/ 
/* #[move_store:		*/

void
move_store(g, number, side, place)
/*
 * Store a new move in the move list
 */
GOGAME *g;
int	number;
int	side;
PLACE	place;
{
    GOMOVE *m = need_move(g, "move");
    char   *error;

    if (!m) return;

    ++(g->gm_movenr);
    m->mv_place = place;
    m->mv_side  = side;
    m->mv_nr    = (number == AUTO_NR) ? g->gm_movenr : number;

    switch (rules_check(g, place, side)) {
	case RULES_LEGAL   : return;
	case RULES_OUTSIDE : error = "off the board";		break;
	case RULES_SUICIDE : error = "suicide";			break;
	case RULES_KO	   : error = "ko capture not allowed";	break;
	case RULES_OCCUPIED: error = "occupied";		break;
	default:	     error = "unknown error";		break;
    }

    /*
     * Emit error info
     */
    message("*** %c[%c%c] (%s) ***\n",
	    (m->mv_side == BLACK) ? 'B' : 'W',
	    P2a(g,place), P2b(g,place),
	    error);

    /*
     * Free the resources
     */
    g->gm_movelast = move_free(m);
}

/* #]move_store:		*/ 
/* #[pass_store:		*/

void
pass_store(g, number, side)
/*
 * Store a "pass"
 */
GOGAME *g;
int	number;
int	side;
{
    GOMOVE *m = need_move(g, "pass");

    if (!m) return;

    ++(g->gm_movenr);
    m->mv_place = MV_PASS;
    m->mv_side  = side;
    m->mv_nr    = (number == AUTO_NR) ? g->gm_movenr : number;
}

/* #]pass_store:		*/ 
/* #[node_store:		*/

void
node_store(g)
/*
 * Store an SGF node indicator
 */
GOGAME *g;
{
    GOMOVE *m = move_new(g, g->gm_movelast);

    if (!m) return;

    g->gm_movelast = m;
}

/* #]node_store:		*/ 
/* #[time_store:		*/

void
time_store(g, side, t)
/*
 * Store time left information
 */
GOGAME *g;
int	side;
char   *t;
{
    GOMOVE *m = need_move(g, "time");

    if (!m) return;

    if (side == BLACK)	m->mv_btime = strdup(t);
    else		m->mv_wtime = strdup(t);
}

/* #]time_store:		*/ 
/* #[letter_store:		*/

void
letter_store(g, place, letter)
/*
 * Add a letter to a position.
 * The destination should be empty (I think).
 * The order of the characters should be maintained,
 * since that decides which character to display (a,b,...).
 */
GOGAME *g;
PLACE	place;
LETTER	letter;
{
    GOMOVE *m = need_move(g, "letter");
    CHAIN  *c, *d;

    if (!m || (c = chain_alloc()) == NULL) return;

    if (m->mv_letter) {
	/*
	 * Seek to the end of the list
	 */
	for (d = m->mv_letter; d && d->ch_next; d = d->ch_next);

	d->ch_next   = c;
	c->ch_letter = (letter == AUTO_LETTER) ? (d->ch_letter + 1) : letter;
    } else {
	m->mv_letter = c;
	c->ch_letter = (letter == AUTO_LETTER) ? 'a' : letter;
    }
    c->ch_place  = place;
}

/* #]letter_store:		*/ 
/* #[mark_store:		*/

void
mark_store(g, place, mark)
/*
 * Add a mark to a position, automatic (*mark == AUTOMARK), or forced.
 * The destination might be occupied (I think).
 * The order of the marks is not important but we maintain it anyway.
 */
GOGAME *g;
PLACE	place;
MARK	mark;
{
    GOMOVE *m = need_move(g, "mark");
    CHAIN  *c, *d;
    MARK    themark;

    if (!m || (c = chain_alloc()) == NULL) return;

    if (m->mv_mark) {
	/*
	 * Seek to the end of the list
	 */
	for (d = m->mv_mark; d && d->ch_next; d = d->ch_next);

	d->ch_next = c;
	if (!*mark) strcpy(c->ch_mark, d->ch_mark);
    } else {
	m->mv_mark = c;
	if (!mark || !*mark) sprintf(c->ch_mark, "%c", TRIANGLE_FILLED);
    }
    if (mark && *mark) {
	strncpy (themark, mark, sizeof(MARK)-1);
	themark[sizeof(MARK)-1] = 0;
	strcpy (c->ch_mark, themark);
    }
    c->ch_place    = place;
    c->ch_oldstone = NODE(g,place)->gn_stone;
}

/* #]mark_store:		*/ 
/* #[view_store:		*/

void
view_store(g, i, j)
/*
 * Store a corner of the view
 */
GOGAME *g;
PLACE	i;
PLACE	j;
{
    if (i < 0 || i >= g->gm_sz ||
	j < 0 || j >= g->gm_sz) {
	fprintf(stderr, "Illegal point: VW[%c%c]", i+'a', j+'a');
    }
	 if (!g->gm_vw[2]) g->gm_vw[2] = i;
    else if (!g->gm_vw[0]) g->gm_vw[0] = i;
	 if (!g->gm_vw[3]) g->gm_vw[3] = j;
    else if (!g->gm_vw[1]) g->gm_vw[1] = j;
    if (g->gm_vw[0] > g->gm_vw[2]) {
	i = g->gm_vw[0];
	    g->gm_vw[0] = g->gm_vw[2];
	    g->gm_vw[2] = i;
    }
    if (g->gm_vw[1] > g->gm_vw[3]) {
	j = g->gm_vw[1];
	    g->gm_vw[1] = g->gm_vw[3];
	    g->gm_vw[3] = j;
    }
}

/* #]view_store:		*/ 
/* #[reg_store:			*/

void
reg_store(g, i, j)
/*
 * Store a corner of the region
 */
GOGAME *g;
PLACE	i;
PLACE	j;
{
	 if (!g->gm_rg[2]) g->gm_rg[2] = i;
    else if (!g->gm_rg[0]) g->gm_rg[0] = i;
	 if (!g->gm_rg[3]) g->gm_rg[3] = j;
    else if (!g->gm_rg[1]) g->gm_rg[1] = j;
    if (g->gm_rg[0] > g->gm_rg[2]) {
	i = g->gm_rg[0];
	    g->gm_rg[0] = g->gm_rg[2];
	    g->gm_rg[2] = i;
    }
    if (g->gm_rg[1] > g->gm_rg[3]) {
	j = g->gm_rg[1];
	    g->gm_rg[1] = g->gm_rg[3];
	    g->gm_rg[3] = j;
    }
}

/* #]reg_store:			*/ 
/* #[rnd_store:			*/

void
rnd_store(g, rnd)
/*
 * Store round of tournament the game was played
 */
GOGAME *g;
int	rnd;
{
    if (!g->gm_ro) g->gm_ro = rnd;
    else
    /*
     * Emit error info
     */
    message("*** More than one ROund specifier ***\n");
}

/* #]rnd_store:			*/ 
/* #[addstone_store:		*/

void
addstone_store(g, side, place)
/*
 * Add a stone to a position
 * Don't check the Go rules (!)
 * The destination might already be occupied.
 */
GOGAME *g;
int	side;		/* BLACK, or WHITE */
PLACE	place;
{
    GOMOVE *m = need_move(g, "stone");
    CHAIN  *c, *d;

    if (!m || (c = chain_alloc()) == NULL) return;

    if (m->mv_addstone) {
	/*
	 * Seek to the end of the list
	 */
	for (d = m->mv_addstone; d && d->ch_next; d = d->ch_next);

	d->ch_next = c;
    } else {
	m->mv_addstone = c;
    }
    c->ch_oldstone = NODE(g,place)->gn_stone;
    c->ch_newstone = move_new(g, c->ch_newstone);
    c->ch_newstone->mv_side  = side;
    c->ch_newstone->mv_nr    = 0;
    c->ch_newstone->mv_place = place;
    c->ch_place		     = place;

    /*
     * Update the board
     */
    NODE(g,place)->gn_stone = c->ch_newstone;
}

/* #]addstone_store:		*/ 
/* #[delstone_store:		*/

void
delstone_store(g, place)
/*
 * Remove a stone from a position
 */
GOGAME *g;
PLACE	place;
{
    GOMOVE *m = need_move(g, "removal");
    CHAIN  *c, *d;

    if (!m || (c = chain_alloc()) == NULL) return;

    if (NODE(g,place)->gn_stone == NULL) {
	int ok = 0;

	/*
	 * The place is not occupied. We silently allow the user
	 * to throw away the same stone twice, but we dont allow
	 * him to throw away an empty point from the start.
	 */
	for (d = m->mv_delstone; d; d = d->ch_next)
	    if (d->ch_oldstone->mv_place == place) ok = 1;

	if (!ok) message("AE[%c%c]\t*** (not occupied)\n",
			 P2a(g,place),
			 P2b(g,place));
	mp_free(cpool, c);
	return;
    }

    if (m->mv_delstone) {
	/*
	 * Seek to the end of the list
	 */
	for (d = m->mv_delstone; d && d->ch_next; d = d->ch_next);

	d->ch_next = c;
    } else {
	m->mv_delstone = c;
    }
    c->ch_oldstone = NODE(g,place)->gn_stone;
    c->ch_newstone = NULL;
    c->ch_place	   = place;

    /*
     * Update the board
     */
    NODE(g,place)->gn_stone = NULL;
}

/* #]delstone_store:		*/ 
/* #[gamename_store:		*/

void
gamename_store(g, name)
/*
 * Store the GameName of the game
 */
GOGAME *g;
char   *name;
{
    if (name && *name) g->gm_gn = strdup(name);
}

/* #]gamename_store:		*/ 
/* #[name_store:		*/

void
name_store(g, name)
/*
 * Store the Name of the game/variation
 */
GOGAME *g;
char *name;
{
    if (name && *name) g->gm_n = strdup(name);
}

/* #]name_store:		*/ 
/* #[user_store:		*/

void
user_store(g, msg)
/*
 * Store some USer data on the game
 */
GOGAME *g;
char *msg;
{
    if (msg && *msg) g->gm_us = strdup(msg);
}

/* #]user_store:		*/ 
/* #[size_store:		*/

void
size_store(g, size)
/*
 * Store the boardsize of a game/variation
 */
GOGAME *g;
int	size;
{
    if (2 >= size || size > MAXSIZE) {
	message("*** %d is an illegal board size ***\n", size);
	size = MAXSIZE;
	message("*** Substituted %d as board size ***\n", size);
    }

    if (size == g->gm_sz) return;

    if (g->gm_move && IS_MOVE(g->gm_move)) {
	/*
	 * We already received moves...
	 */
	message("*** resized board caused information loss ***\n");
    }

    /*
     * Free previous resources
     */
    if (g->gm_goban->gb_node) free(g->gm_goban->gb_node);
    if (g->gm_goban) free(g->gm_goban);
    if ((g->gm_goban = goban_new(g, size)) == NULL) {
	g->gm_sz = 0;
	return;
    }
    g->gm_sz = size;
}

/* #]size_store:		*/ 
/* #[capture_store:		*/

void
capture_store(g, gonode)
/*
 * Add the stone at gonode to the captured stone list
 */
GOGAME *g;
GONODE *gonode;
{
    GOMOVE *m = need_move(g, "prisoner");
    CHAIN  *c;

    if (!m || (c = chain_alloc()) == NULL) return;

    c->ch_oldstone = gonode->gn_stone;
    c->ch_place    = gonode->gn_stone->mv_place;

    /*
     * Insert the stone
     */
    c->ch_next	   = m->mv_captured;
    m->mv_captured = c;
}

/* #]capture_store:		*/ 
/* #[need_move:			*/

static GOMOVE *
need_move(g, why)
/*
 * Check the current move, if it's empty insert a new move.
 */
GOGAME *g;
char   *why;
{
    if (g->gm_movelast == NULL) {
	message("*** isolated info (\"%s\"), I allocated storage ***\n", why);
	g->gm_movelast = move_new(g, g->gm_movelast);
    }
    return g->gm_movelast;
}

/* #]need_move:			*/ 
/* #[push_var:			*/

GOGAME *
push_var(gogame)
/*
			#[explanation:
 */
#if 0
		Global structure of moves and diagrams
		--------------------------------------

	game	( move_1 move_2 move_3 ...... move_n-1 move_n )
			|
	var1		( move_1 ...... move_n )->gm_parent
	var2		( move_1 ...... move_n )->gm_parent
				|
	var2_1			( move_1 ...... move_n )--> var2

	We maintain three pointers to indicate these relations:

	* gm_parent	points to current game/variation
	* gm_next	points to next game/variation on same level
	* gm_prev	points to prev game/variation on same level

	Explanation:

	var1->gm_next   = var2	var2->gm_next   = NULL
	var1->gm_prev   = NULL	var2->gm_prev   = var1
	var1->gm_parent = game	var2->gm_parent = game

				var2_1->gm_next   = NULL
				var2_1->gm_prev   = NULL
				var2_1->gm_parent = var2
#endif
/*
			#]explanation: 
 */
GOGAME *gogame;
{
    GOMOVE *m = need_move(gogame, "variation");
    GOGAME *new;
    GOGAME *p, *q = NULL;

    if (!m) return NULL;

    /*
     * First, find an empty slot for this variation
     */
    p = m->mv_variation;
    while (p) { q = p; p = p->gm_next; }

    /*
     * Next, store the new variation
     */
    if ((new = game_alloc(gogame)) == NULL) return NULL;
    new->gm_prev	= q;
    new->gm_next	= NULL;
    new->gm_parent	= gogame;
    if (q) q->gm_next	= new;
    if (q) new->gm_gid	= q->gm_gid + 1;
    if (m->mv_variation == NULL)
	m->mv_variation = new;

    return new;
}

/* #]push_var:			*/ 
/* #[pop_var:			*/

GOGAME *
pop_var(g)
/*
 * Restore the board status how we found it
 */
GOGAME *g;
{
    move_mv_allback(g);

    return g->gm_parent;
}

/* #]pop_var:			*/ 


/* #[move_current:		*/

GOMOVE *
move_current(g)
GOGAME 	*g;
{
    if (g == NULL) {
	message("Game pointer null in move_current()\n");
	return NULL;
    }

    /*
     * Return the last played move
     */
    return g->gm_movelast;
}

/* #]move_current:		*/ 
/* #[move_mv_forw:		*/

GOMOVE *
move_mv_forw(g)
/*
 * Move through the game --- move_mv_forw()
 * Show the next move played (if any)
 */
GOGAME *g;
{
    GOMOVE *m = move_current(g);
    CHAIN  *c;

    /*
     * Fetch the next move
     */
    m = (m == NULL) ? g->gm_move : m->mv_next;

    if (m == NULL) {
	/*
	 * No next move, we can't go forward
	 */
	return NULL;
    }

    /*
     * Place the stone
     */
    if (IS_MOVE(m)) {
	NODE(g,m->mv_place)->gn_stone = m;
    }

    /*
     * Remove the captured stones
     */
    for (c = m->mv_captured; c; c = c->ch_next) {
	NODE(g,c->ch_oldstone->mv_place)->gn_stone = NULL;
    }

    /*
     * Remove the deleted stones
     */
    for (c = m->mv_delstone; c; c = c->ch_next) {
	NODE(g,c->ch_oldstone->mv_place)->gn_stone = NULL;
    }

    /*
     * Add the added stones
     */
    for (c = m->mv_addstone; c; c = c->ch_next) {
	NODE(g,c->ch_newstone->mv_place)->gn_stone = c->ch_newstone;
    }

    /*
     * Update the current move
     */
    g->gm_movelast = m;

    return g->gm_movelast;
}

/* #]move_mv_forw:		*/ 
/* #[move_mv_back:		*/

GOMOVE *
move_mv_back(g)
/*
 * Move through the game --- move_mv_back()
 * Remove the last move played (if any)
 */
GOGAME *g;
{
    GOMOVE *m = move_current(g);
    CHAIN  *c;

    if (m == NULL) {
	/*
	 * No current move, we can't go back
	 */
	return NULL;
    }

    /*
     * Take away the stone
     */
    if (IS_MOVE(m)) {
	NODE(g,m->mv_place)->gn_stone = NULL;
    }

    /*
     * Put the captured stones back
     */
    for (c = m->mv_captured; c; c = c->ch_next) {
	NODE(g,c->ch_oldstone->mv_place)->gn_stone = c->ch_oldstone;
    }

    /*
     * Put the deleted stones back
     */
    for (c = m->mv_delstone; c; c = c->ch_next) {
	NODE(g,c->ch_oldstone->mv_place)->gn_stone = c->ch_oldstone;
    }

    /*
     * Remove the added stones
     */
    for (c = m->mv_addstone; c; c = c->ch_next) {
	NODE(g,c->ch_newstone->mv_place)->gn_stone = c->ch_oldstone;
    }

    /*
     * Update the current move
     */
    g->gm_movelast = m->mv_prev;

    return g->gm_movelast;
}

/* #]move_mv_back:		*/ 
/* #[move_mv_allback:		*/

GOMOVE *
move_mv_allback(g)
/*
 * Move through the game --- move_mv_allback()
 * Remove all moves (if any)
 */
GOGAME *g;
{
    while (move_mv_back(g));

    return g->gm_movelast;
}

/* #]move_mv_allback:		*/ 
/* #[move_mv_allforw:		*/

GOMOVE *
move_mv_allforw(g)
/*
 * Move through the game --- move_mv_allforw()
 * Show all moves (if any)
 */
GOGAME *g;
{
    while (move_mv_forw(g));

    return g->gm_movelast;
}

/* #]move_mv_allforw:		*/ 

/* #[game_name:			*/

char *
game_name(g)
/*
 * Return the name of the game or something else sensible
 */
GOGAME *g;
{
    if (g              == NULL) return "game nil";
    if (g->gm_gn != NULL) return g->gm_gn;
    if (g->gm_n     != NULL) return g->gm_n;
    return "";
}

/* #]game_name:			*/ 
