/* #[info:		*/
/************************************************************************
*									*
*		  ,,	,,		    ,				*
*		  ||  ' ||		   ||				*
*		  || \\ ||/|,  _-_  ,._-_ =||= '\\/\\			*
*		  || || || || || \\  ||	   ||	|| ;'			*
*		  || || || |' ||/    ||	   ||	||/			*
*		  \\ \\ \\/   \\,/   \\,   \\,	|/			*
*					       (			*
*						-_-			*
*									*
*		  Copyright (c) 1993  Jan van der Steen			*
*	       Centre for Mathematics and Computer Science		*
*		       Amsterdam, the Netherlands			*
*									*
*-----------------------------------------------------------------------*
*   You are not allowed to use this software in commercial software.	*
*   You are allowed to use this software in free software under the	*
*   condition that you maintain this integral message in the source.	*
*-----------------------------------------------------------------------*
* File	  : liberty.c							*
* Purpose : Read a Liberty Go game					*
* Version : 1.13							*
* Modified: 03 May 1995 10:28:07					*
* Author  : Jan van der Steen (jansteen@cwi.nl)				*
************************************************************************/
/* #]info:		*/ 
/* #[include:		*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "sgf2misc.h"
#include "gogame.h"
#include "tools.h"
#include "liberty.h"

#ifdef __STRICT_ANSI__
char * strdup           (const char *);
int    strncasecmp      (const char *s1, const char *s2, size_t n);
int    isascii          (int c);
#endif

/* #]include:		*/ 
/* #[define:		*/

#define MAXBUFFER 10000L	/* fread() IO buffer */
#define MAXBUF	  1024		/* fgets() IO buffer */

#define LIB2P(p)	(LIB_X(p) + (LIB_Y(p) * g->gm_sz))

#if defined (ATARI)
#define RAWREAD	"rb"
#else
#define RAWREAD "r"
#endif

/* #]define:		*/ 
/* #[extern:		*/

extern char *	program;		/* program name		*/

/* #]extern:		*/ 
/* #[static:		*/

static	unsigned char *	LibBuffer;	/* buffer of lib file	*/
static	unsigned long	BufferSize;	/* size of lib buffer	*/
static	int		readmagic;
static	int		readsize;
static	int		n_dia  = 0;
static	int		n_move = 0;

/* #]static:		*/ 
/* #[prototype:		*/

#ifdef __STDC__
#       define  PROTO(s) s
#else
#       define  PROTO(s) ()
#endif

static	void		readgame	PROTO((GOGAME *g, char *filename));
static	void		readtxt		PROTO((GOGAME *g, char *filename));
static	unsigned long	readlong	PROTO((FILE *fp));

static	void		lib2txt		PROTO((char *path , char *suffix));
static	void		getmagic	PROTO((FILE *fp , char *filename));
static	void		getboardsize	PROTO((GOGAME *g, FILE *fp));
static	void		getmoves	PROTO((FILE *fp));
static	void		getcore		PROTO((void));
static	void		lib_store	PROTO((GOGAME *g));

#undef PROTO

/* #]prototype:		*/ 

/* #[lib_store:		*/

static void
lib_store(gogame)
/*
 * Replay the stack and store it...
 */
GOGAME *gogame;
{
    GOGAME	  *	g = gogame;
    unsigned char *	b;
    int			india = 0;
    int			color = BLACK;

    for (b = LibBuffer; BufferSize > 0; BufferSize--) {
	unsigned long p[1];

/* #[fetch_long:	*/
	/*
	 * Get Liberty long
	 */
	p[0]  = (((unsigned long) *b++) << 24);
	p[0] += (((unsigned long) *b++) << 16);
	p[0] += (((unsigned long) *b++) <<  8);
	p[0] += (((unsigned long) *b++)      );

/* #]fetch_long:	*/ 
/* #[variation_end:	*/

	/*
	 * End of variation?!
	 */
	if (india && (
	    (LIB_N(p) == LIB_FIG) ||
	    (LIB_N(p) == LIB_DIA && LIB_STATE(p) == LIB_DIAGRAM))) {

	    GOGAME *parent = pop_var(g);

	    /*
	     * The next can't happen in a Liberty game
	     */
	    if (parent == (GOGAME *) 0) goto error_parse;
	    g = parent;

	    india = 0;
	}

/* #]variation_end:	*/ 
/* #[variation_start:	*/

	/*
	 * A new diagram?!
	 */
	if (LIB_N(p) == LIB_DIA && LIB_STATE(p) == LIB_DIAGRAM) {
	    char var_name[128];

	    g = push_var(g);

	    /*
	     * Store a node and name for the variation
	     */
	    node_store(g);
	    sprintf(var_name, "Diagram %d", ++n_dia);
	    name_store(g, var_name);

	    /*
	     * Finally, increment the counter (skip two more)
	     */
	    BufferSize -= 2; b += 8;
	    india = 1;
	    continue;
	}

/* #]variation_start:	*/ 
/* #[events:		*/

	switch(LIB_STATE(p)) {
	    case LIB_STONE:
		addstone_store(g,
			(LIB_STAT(p) == LIB_WHITE) ? WHITE : BLACK, LIB2P(p));
		break;
	    case LIB_NUMSTONE:
		/*
		 * Nice moment to store a node as well
		 */
		node_store(g);
		color = (LIB_STAT(p) == LIB_WHITE) ? WHITE : BLACK;
		move_store(g, LIB_NUM(p), color, LIB2P(p));
		if (LIB_N(p) == LIB_FIG) n_move++;
		break;
	    case LIB_PASS:
	    case LIB_TENUKI:
		color = (color == BLACK) ? WHITE : BLACK;
		pass_store(g, LIB_NUM(p), color);
		break;
	    case LIB_MARKSTONE:
		{
		    MARK mark;
		    
		    switch (LIB_MARK(p)) {
			case LIB_PLUS	: strcpy (mark, "+"); break;
			case LIB_MINUS	: strcpy (mark, "-"); break;
			default		: sprintf(mark, "%c", TRIANGLE_EMPTY);
		    }
		    mark_store(g, LIB2P(p), mark);
		}
		break;
	    case LIB_CHARACTER:
		letter_store(g, LIB2P(p), LIB_CHAR(p) + 'A' - 1);
		break;

	    /*
	     * Unbound for the moment
	     */
	    case LIB_HIT:
	    case LIB_HITNUM:
	    case LIB_HITSTONE:
	    case LIB_CONTINUE:
	    case LIB_DIAGRAM:
		break;

	    default:
		message("lib_store(): Undefined case: %d\n", LIB_STATE(p));
	}
/* #]events:		*/ 
    }
error_parse:
    return;
}

/* #]lib_store:		*/ 
/* #[readtxt:		*/

static void
readtxt(g, filename)
GOGAME * g;
char   * filename;
{
    FILE *	fp;
    char	txtfile[MAXBUF];
    char	line   [MAXBUF];
    static int	n = 0;

    strcpy(txtfile, filename);
    lib2txt(txtfile, "txt");
    if ((fp = fopen(txtfile, "r")) == (FILE *) 0) {
	message("\n%s: Couldn't open \"%s\"", program, txtfile);
	return;
    }
    while (fgets(line, MAXBUF, fp) != NULL) {
	char *s;

	if ((s = strchr(line, '\n')) != (char *) 0) *s = 0;
	if ((s = strchr(line, '\r')) != (char *) 0) *s = 0;

	switch(n++) {
	case 0 : g->gm_pb    = strdup(line);	break;
	case 1 : g->gm_pw    = strdup(line);	break;
	case 2 : g->gm_ha = strdup(line);	break;
	case 3 : g->gm_km     = strdup(line);	break;
	case 4 : g->gm_dt     = strdup(line);	break;
	case 5 : g->gm_pc    = strdup(line);	break;
	case 6 : g->gm_re   = strdup(line);	break;
#if 0
	case 7 : if (*line) text_store(g, line);	break;
	case 8 : if (*line) text_store(g, line);	break;
	case 9 : if (*line) text_store(g, line);	break;
#endif
	default: text_store(g, line);
	}
    }
    fclose(fp);
}

/* #]readtxt:		*/ 
/* #[readgame:		*/

static void
readgame(g, filename)
GOGAME * g;
char   * filename;
{
    FILE *fp;
    char binfile[MAXBUF];

    strcpy(binfile, filename);
    lib2txt(binfile, "lib");
    if ((fp = fopen(binfile, RAWREAD)) == NULL ) {
	message("%s: Couldn't open \"%s\"\n", program, binfile);
	exit(1);
    }
    getcore	 ();
    getmagic	 (fp, binfile);
    getboardsize (g, fp);
    getmoves	 (fp);
    fclose	 (fp);
    lib_store	 (g);
    free	 (LibBuffer);
}

/* #]readgame:		*/ 
/* #[readlong:		*/

#define BYTE(n)	((int) ((n) & 0x7f) | (((n) & 0x80) ? 0x80 : 0x00))

static unsigned long
readlong(fp)
/*
 * Read a BIGENDIAN long from stream fp
 */
FILE *fp;
{
    char tmp[4];

    if (fread(tmp, 1, 4L, fp) != 4) {
	message("%s: Read error\n", program);
	exit (1);
    }
    return (((((BYTE(tmp[0])  << 8) +
		BYTE(tmp[1])) << 8) +
		BYTE(tmp[2])) << 8) +
		BYTE(tmp[3]);
}

/* #]readlong:		*/ 
/* #[read_liberty:	*/

GOGAME *
read_liberty(filename)
char *filename;
{
    GOGAME *g = game_alloc((GOGAME *) 0);

    if (filename == NULL) {
	fprintf(stderr,
		"%s: Can't read liberty format from standard input\n",
		program);
	exit(1);
    }
    node_store(g);
    readtxt	 (g, filename);
    readgame	 (g, filename);
    message	 ("[%d moves, %d variations]\n", n_move, n_dia);

    return g;
}

/* #]read_liberty:	*/ 

/* #[lib2txt:		*/

static void
lib2txt(path, suffix)
char *path;
char *suffix;
/*
 * Convert a filename to a name ending in "suffix".
 * Path should have enough space to contain the extra characters.
 */
{
    int i = (int) strlen(path);

    if (!i) return;				/* empty path */
    else i--;

    /*
     * find last backslash...
     */
    while (i && path[i] != DIRSEP) i--;

    /*
     * find dot, or make one...
     */
    while (path[i] && path[i] != '.') i++;
    if (path[i] != '.') strcat(path+i, ".");
    /*
     * append suffix...
     */
    strcpy(path+i+1, suffix);
}

/* #]lib2txt:		*/ 
/* #[getmagic:		*/

static void
getmagic(fp, filename)
FILE *fp;
char *filename;
{
    char magic[4];

    if ( fread(magic, 1, 4L, fp) != 4 ) {
	message("%s: Read error\n", program);
	exit(1);
    }
    if ( strncmp(magic, "ISHI", 4) ) {
	message("\n%s: \"%s\" could be a non-Liberty file",
		program, filename);
	fseek(fp, 0L, 0);
	readmagic = 0;
    }
    else readmagic = 4;
}

/* #]getmagic:		*/ 
/* #[getboardsize:	*/

static void
getboardsize(g, fp)
GOGAME * g;
FILE   * fp;
{
    unsigned long lsize = readlong(fp);

    if ( (int)(lsize & STATEMASK) == LIB_BOARDSIZE ) {
	g->gm_sz = 2 * ((int)(lsize >> NUMSHIFT) & NUMMASK) + 1;
	readsize = 4;
	message("[Board size = %02dx%02d]", g->gm_sz, g->gm_sz);
    }
    else {
	g->gm_sz = MAXSIZE;
	fseek(fp, -4L, 1);
	readsize = 0;
	message("[Assumed board size = %02dx%02d]", g->gm_sz, g->gm_sz);
    }
    if (g->gm_sz != MAXSIZE) size_store(g, g->gm_sz);
}

/* #]getboardsize:	*/ 
/* #[getmoves:		*/

static void
getmoves(fp)
FILE *fp;
{
    BufferSize = fread(LibBuffer, 1, MAXBUFFER, fp);
    message("[%ld bytes]", BufferSize+readmagic+readsize);
    BufferSize >>= 2;	/* number of longs */
}

/* #]getmoves:		*/ 
/* #[getcore:		*/

static void
getcore()
{
    LibBuffer = (unsigned char *) malloc(MAXBUFFER);
    if ( !LibBuffer ) {
	message("%s: Out of memory in getcore()\n", program);
	exit(1);
    }
}

/* #]getcore:		*/ 
