/* #[info:		*/
/************************************************************************
*									*
*	                                                ,,		*
*	                                                ||		*
*	      \\/\\/\\  _-_  \\/\\/\\ -_-_   /'\\  /'\\ ||		*
*	      || || || || \\ || || || || \\ || || || || ||		*
*	      || || || ||/   || || || || || || || || || ||		*
*	      \\ \\ \\ \\,/  \\ \\ \\ ||-'  \\,/  \\,/  \\		*
*	                              |/                  		*
*	                              '                   		*
*									*
*		 Copyright (c) 1993  Jan van der Steen			*
*	      Centre for Mathematics and Computer Science		*
*		      Amsterdam, the Netherlands			*
*									*
*-----------------------------------------------------------------------*
*   You are not allowed to use this software in commercial software.	*
*   You are allowed to use this software in free software under the	*
*   condition that you maintain this integral message in the source.	*
*-----------------------------------------------------------------------*
* File    : mempool.c							*
* Purpose : ___________________________________________________________	*
* Version : 1.5								*
* Modified: 19 Oct 1994 14:04:24					*
* Author  : Jan van der Steen (jansteen@cwi.nl)				*
************************************************************************/
/* #]info:		*/ 
/* #[include:		*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef ATARI
#include <malloc.h>
#endif

#include "mempool.h"

/* #]include:		*/ 
/* #[define:		*/

#define BLKSZ(mp)	(sizeof(MEM) + mp->mp_sz)

/* #]define:		*/ 
/* #[prototype:		*/

#ifdef __STDC__
#	define	PROTO(s) s
#else
#	define	PROTO(s) ()
#endif

static	MP *	mp_new		PROTO((MP *mp, long n, long dn, long sz));
static	MEM *	mp_insert	PROTO((MP *mp, MP *mq, long n));
static	int	mp_check	PROTO((MP *mp, MEM *m));

#undef PROTO

/* #]prototype:		*/ 

/* #[mp_new:		*/

static MP *
mp_new(root, n, dn, sz)
/*
 * Initialise a new memory pool of n elements
 */
MP * root;
long n;
long dn;
long sz;
{
    MP *parent, *new;

    if (!sz || (new = calloc(1, sizeof(MP))) == NULL) return NULL;
    new->mp_n  = n;
    new->mp_dn = dn;
    new->mp_sz = sz;

    if (root && (parent = root->mp_last) != NULL) {
	new->mp_n += parent->mp_n;
	parent->mp_next = new;
	root->mp_last   = new;
    } else
	new->mp_last    = new;	/* Self */

    /*
     * sizeof(MEM) is our administration,
     * sz is the requested block
     */
    if ((new->mp_mem = calloc(n, sizeof(MEM)+sz)) == NULL) {
	free(new);
	return NULL;
    }
    return new;
}

/* #]mp_new:		*/ 
/* #[mp_insert:		*/

static MEM *
mp_insert(mp, mq, n)
/*
 * Insert n free blocks, return address of first element
 */
MP * mp;	/* Master pool			*/
MP * mq;	/* Pool containing new memory	*/
long n;
{
    char *p = mq->mp_mem;

    /*
     * Firstly, convert the n MEM blocks into a linked list
     */
    while (n-- > 0) {
	((MEM *) p)->mem_next = (MEM *) (p + BLKSZ(mp));
	p += BLKSZ(mp);
    }

    /*
     * Set stopper at the tail of the list
     */
    ((MEM *)(p - BLKSZ(mp)))->mem_next = NULL;

    return (MEM *)(mq->mp_mem);
}

/* #]mp_insert:		*/ 

/* #[mp_init:		*/

MP *
mp_init(n, dn, sz, flags)
/*
 * Initialise a memory pool of n elements with sz bytes each.
 * When reallocation is necessary, it will grow with dn elements.
 */
long n;
long dn;
long sz;
int  flags;
{
    MP *mp;

    n   = (n  == 0) ? MP_N  : n;
    dn  = (dn == 0) ? MP_DN : dn;
    sz += (sz & 0x03) ? (4 - (sz & 0x03)) : 0;	/* 4 byte alignment */
    if (!sz || (mp = mp_new(NULL, n, dn, sz)) == NULL) return NULL;

    mp->mp_free  = mp_insert(mp, mp, mp->mp_n);
    mp->mp_flags = flags;

    return mp;
}

/* #]mp_init:		*/ 
/* #[mp_stat:		*/

MP *
mp_stat(root)
/*
 * Return last allocated memory pool
 */
MP *root;
{
    return (root) ? root->mp_last : NULL;
}

/* #]mp_stat:		*/ 
/* #[mp_alloc:		*/

void *
mp_alloc(mp)
/*
 * Return next free block, realloc block pool if necessary
 */
MP *mp;
{
    MEM *m;

    if (!mp) return NULL;

    if ((m = mp->mp_free) == NULL) {
	/*
	 * Reallocation needed (mp->mp_dn extra elements)
	 */
	MP *new = mp_new(mp, mp->mp_dn, mp->mp_dn, mp->mp_sz);

	if (!new) return NULL;
	m = mp->mp_free = mp_insert(mp, new, mp->mp_dn);

	if (mp->mp_flags & MP_MSG)
	    fprintf(stderr, "MP: Allocation of %ld blocks\n", mp->mp_dn);
    }
    mp->mp_free = mp->mp_free->mem_next;

    return (void *)(((char *) m) + sizeof(MEM));
}

/* #]mp_alloc:		*/ 
/* #[mp_calloc:		*/

void *
mp_calloc(mp)
/*
 * Return cleared block
 */
MP *mp;
{
    void *p = mp_alloc(mp);

    if (p) memset(p, 0, mp->mp_sz);

    return p;
}

/* #]mp_calloc:		*/ 
/* #[mp_free:		*/

void
mp_free(mp, p)
MP   *mp;
void *p;
{
    MEM *m;

    if (!mp || !p) return;

    m = (MEM *)((char *) p - sizeof(MEM));
    if (!(mp->mp_flags & MP_CHK) || mp_check(mp, m)) {
	/*
	 * Move this element to head of the list
	 */
	m->mem_next = mp->mp_free;
	mp->mp_free = m;
    } else {
	/*
	 * Freeing a non-member....
	 */
	fprintf(stderr, "MP: *** Freeing a non-member ***\n");
	/* abort(); */
    }
}

/* #]mp_free:		*/ 
/* #[mp_check:		*/

static int
mp_check(root, m)
/*
 * Check validity of m
 */
MP  *root;
MEM *m;
{
    MP *p;

    for (p = root; p; p = p->mp_next) {
	if ((char *) m >= p->mp_mem &&
	    (char *) m <  p->mp_mem +
		 BLKSZ(root) * ((p == root) ? p->mp_n : p->mp_dn)) return 1;
    }
    return 0;
}

/* #]mp_check:		*/ 
/* #[mp_exit:		*/

void
mp_exit(mp)
/*
 * Release the whole pool
 */
MP *mp;
{
    if (mp) {
	if (mp->mp_next) mp_exit(mp->mp_next);
	if (mp->mp_mem ) free(mp->mp_mem);
	free(mp);
    }
}

/* #]mp_exit:		*/ 

