/* #[info:		*/
/************************************************************************
*									*
*			    /\	/\					*
*		      _	   ||  (  )	      '				*
*	       _-_,  / \\ =||=	 // \\/\\/\\ \\	 _-_,  _-_		*
*	      ||_.  || ||  ||	//  || || || || ||_.  ||		*
*	       ~ || || ||  ||  /(   || || || ||	 ~ || ||		*
*	      ,-_-  \\_-|  \\, {___ \\ \\ \\ \\ ,-_-  \\,/		*
*		     /	\						*
*		    '----`						*
*									*
*		  Copyright (c) 1993  Jan van der Steen			*
*	       Centre for Mathematics and Computer Science		*
*		       Amsterdam, the Netherlands			*
*									*
*-----------------------------------------------------------------------*
*   You are not allowed to use this software in commercial software.	*
*   You are allowed to use this software in free software under the	*
*   condition that you maintain this integral message in the source.	*
*-----------------------------------------------------------------------*
* File	  : sgf2misc.c							*
* Purpose : Read/write Go games in various formats			*
* Version : 1.37							*
* Modified: 03 May 1995 12:06:51					*
* Author  : Jan van der Steen (jansteen@cwi.nl)				*
************************************************************************/
/* #]info:		*/ 
/* #[include:		*/

#include <stdlib.h>
#include <string.h>
#include "sgf2misc.h"
#include "gogame.h"
#include "sgfread.h"
#include "liberty.h"
#include "sfdefs.h"
#include "tools.h"
#include "d_ascii.h"
#include "d_tex.h"
#include "d_pstex.h"
#if !NOCURSES
#include "d_screen.h"
#endif
#include "d_sgf.h"
#include "d_sf.h"
#include "d_db.h"
#include "d_gif.h"
#include "d_html.h"
#include "d_null.h"

#ifdef __STRICT_ANSI__
char * strdup           (const char *);
int    strncasecmp      (const char *s1, const char *s2, size_t n);
int    isascii          (int c);
#endif

#ifdef memcpy
#undef memcpy
#endif

/* #]include:		*/ 
/* #[define:		*/

#define	IS_LETTER	8
#define	IS_MARK		12

/* #]define:		*/ 
/* #[typedef:		*/

#ifdef __STDC__
#	define	PROTO(s) s
#else
#	define	PROTO(s) ()
#endif

typedef struct dev_info {
    int		(*dev_init   ) PROTO((void));
    void	(*dev_exit   ) PROTO((void));
    int		(*dev_alloc  ) PROTO((GOGAME *g));
    void	(*dev_free   ) PROTO((GOGAME *g));
    void	(*dev_onmove ) PROTO((GOGAME *g, int n1,int c1, int n2,int c2));
    void	(*dev_onstone) PROTO((GOGAME *g, int n, int c, PLACE p));
    void	(*dev_pass   ) PROTO((GOGAME *g, int n, int c));
    void	(*dev_time   ) PROTO((GOGAME *g, char *btime, char *wtime));
    int		(*dev_emit   ) PROTO((GOGAME *g, GOMOVE *m1, GOMOVE *m2,
						 GOMOVE *t1, GOMOVE *t2));
} PRN_DEVICE;

#undef PROTO

/* #]typedef:		*/ 
/* #[static:		*/

static PRN_DEVICE prn_devices[] = {
	/*#[PSTeX:	*/
	{
		PSTeX_init, PSTeX_exit, PSTeX_alloc, PSTeX_free,
		PSTeX_onmove, PSTeX_onstone, PSTeX_pass, PSTeX_time, 
    	    	PSTeX_emit,
	},
	/*#]PSTeX:	*/ 
	/*#[TeX:	*/
	{
		TeX_init, TeX_exit, TeX_alloc, TeX_free,
		TeX_onmove, TeX_onstone, TeX_pass, TeX_time,
    	    	TeX_emit,
	},
	/*#]TeX:	*/ 
	/*#[ASCII:	*/
	{
		ASCII_init, ASCII_exit, ASCII_alloc, ASCII_free,
		ASCII_onmove, ASCII_onstone, ASCII_pass, ASCII_time, 
    	    	ASCII_emit,
	},
	/*#]ASCII:	*/ 
	/*#[screen:	*/
	{
#if !NOCURSES
		screen_init, screen_exit, screen_alloc, screen_free,
		screen_onmove, screen_onstone, screen_pass, screen_time, 
    	    	screen_emit,
#else
		null_init, null_exit, null_alloc, null_free,
		null_onmove, null_onstone, null_pass, null_time,
    	    	null_emit,
#endif
	},
	/*#]screen:	*/ 
	/*#[GIF:	*/
	{
		GIF_init, GIF_exit, GIF_alloc, GIF_free,
		GIF_onmove, GIF_onstone, GIF_pass, GIF_time, 
    	    	GIF_emit,
	},
	/*#]GIF:	*/ 
	/*#[HTML:	*/
	{
		HTML_init, HTML_exit, HTML_alloc, HTML_free,
		HTML_onmove, HTML_onstone, HTML_pass, HTML_time, 
    	    	HTML_emit,
	},
	/*#]HTML:	*/ 
	/*#[null:	*/
	{
		null_init, null_exit, null_alloc, null_free,
		null_onmove, null_onstone, null_pass, null_time, 
    	    	null_emit,
	}
	/*#]null:	*/ 
};

/* #]static:		*/ 
/* #[global:		*/

char *	      program   = "sgf2misc";	/* Program name			*/
int	      file_fmt  = FMT_SGF;	/* Input format (SGF, LIB, STD)	*/
int	      prn_dev   = DEVICE_PSTEX;	/* PSTEX,TEX,ASCII,SCREEN,DB,NUL*/
int	      paper_size= PAPER_DEFAULT;/* PAPER_{DEFAULT,LETTER,A4WIDE,A4} */
FILE *	      fw;			/* Output channel for text	*/
int	      verbose   = 0;		/* Verbose mode			*/
int	      caching   = 0;		/* Caching mode			*/
int	      textout   = 1;		/* Emit the text too		*/
int	      statout   = 0;		/* Emit the kibitz stats	*/
int	      coords    = 1;		/* Print coordinates		*/
int	      labels    = 1;		/* Label with move numbers	*/
int	      kibitz    = 0;		/* Interpret comments as kibitz	*/
int	      flatten   = 0;		/* Apply the flatten conversion	*/
int	      sequent   = 0;		/* Apply the sequent conversion	*/
int	      zhuge     = 0;		/* Apply the zhuge conversions	*/
int	      longkey   = 1;		/* Long or short SGF keys	*/
int	      hold	= 0;		/* Don't remove ppm script	*/
int	      figlist[MAXNUM];		/* The figure  list		*/
int	      dialist[MAXNUM];		/* The diagram list		*/
unsigned      maxmoves	= ~0; 	    	/* Maximum moves per figure 	*/
EPSFINFO      epsf;			/* The epsf filename/directory	*/
int	      optind    = 1;		/* index into argv vector	*/

/*
 * GIF device
 */
char *		ppm    = PBMFNT;		/* Location of ppm font	*/
char *		pbmbin = PBMBIN;		/* Location of netpbm	*/
char *		rgb_b  = "rgb\\:00/00/00";	/* Black		*/
char *		rgb_w  = "rgb\\:ff/ff/ff";	/* White		*/
char *		rgb_g  = "rgb\\:ff/e1/4b";	/* Board  : Yellow	*/
char *		rgb_o  = "rgb\\:ff/e1/4b";	/* Outside: Yellow	*/
char *		rgb_r  = "rgb\\:ff/ff/7f";	/* Region : Yellow	*/
int		fntsz  = PBMSZ;			/* Fontsize		*/

/* #]global:		*/ 
/* #[prototype:		*/

#ifdef __STDC__
#	define	PROTO(s) s
#else
#	define	PROTO(s) ()
#endif

	int	main		PROTO((int argc, char **argv));
static	void	rnd2flat	PROTO((GOGAME *g));
static	void	seq2rnd		PROTO((GOGAME *g));
static	void	board_cont	PROTO((GOGAME *g));
static	void	board_restore	PROTO((GOGAME *g, GOMOVE *m1, GOMOVE *m2));
static	char *	epsf_filename	PROTO((char *fullpathname));
static	GOMOVE *game_print	PROTO((GOGAME *g, PRN_DEVICE *dev,
				       GOMOVE *m, int *f, int *d));
static	void	game_emit	PROTO((GOGAME *g, PRN_DEVICE *dev));

#undef PROTO

/* #]prototype:		*/ 

/* #[main:		*/

int
main(argc, argv)
int	argc;
char **	argv;
{
    GOGAME * g = NULL;
    char   * filename;
    char   * format = "Undefined";

#if defined(ATARI)
    argv[0] = program;
#else
    program = argv[0];
#endif

    /*
     * Process the options
     */
    fw = stdout;
    getoptions(argc, argv);

    /*
     * Set some filenames
     */
    if (optind < argc) {
	filename	= argv[optind];
	epsf.eps_name	= epsf_filename(filename);
    } else {
	filename	= NULL;
	epsf.eps_name	= epsf_filename("stdin");
    }

    /*
     * Read the Go game
     */
    switch (file_fmt) {
	case FMT_LIB: format = "Liberty"; g = read_liberty (filename); break;
	case FMT_STD: format =      "SF"; g = read_standard(filename); break;
	case FMT_SGF: format =     "SGF"; g = sgf_read     (filename); break;
    }

    /*
     * We exit when no game read, except when writing a database
     */
    if (g == NULL && prn_dev != DEVICE_DB) {
	fprintf(stderr, "%s: \"%s\" is a bad %s file\n",
		program, filename, format);
	exit(EXIT_FAILURE);
    }

    /*
     * Set real epsf size...
     */
    epsf.eps_width = epsf.eps_size + ((coords) ? 2*epsf.eps_size/g->gm_sz : 0);

    /*
     * Convert the game
     */
    if (sequent) seq2rnd (g);
    if (flatten) rnd2flat(g);

    /*
     * Emit or print the game
     */
    switch (prn_dev) {
	case DEVICE_SGF	: game2sgf(g);		 exit(EXIT_SUCCESS);
	case DEVICE_SF	: game2sf (g);		 exit(EXIT_SUCCESS);
	case DEVICE_DB	: game2db (g, filename); exit(EXIT_SUCCESS);
	default		: game_emit(g, prn_devices+prn_dev);
    }

    exit(EXIT_SUCCESS);
}

/* #]main:		*/
/* #[epsf_filename:	*/ 

static char *
epsf_filename(fullpathname)
/*
 * Look for suffix and delete it.
 * Next, return the basename of the file.
 */
char *fullpathname;
{
    char *s = strdup(fullpathname);
    char *t = s;

    /*
     * First delete the suffix
     */
    while (*s) s++;
    while (s > t && *s != DIRSEP && *s != '.') s--;
    /*
     * We *only* want to remove the suffix seperator
     */
    if (*s == '.') *s = 0;

    /*
     * Next, return the basename
     */
    while (s > t && *s != DIRSEP) s--;
    if (*s == DIRSEP) s++;

    return s;
}

/* #]epsf_filename:	*/ 

/* #[con_cat:		*/

/*
 * When we need more ko/connect message space we allocate
 * it in chunks of CONNECTCHUNK bytes
 */
#define	CONNECTCHUNK	128

char *
con_cat(s)
/*
 * s != (char *) 0	Append s to connect buffer
 * s == (char *) 0	Flush the connect buffer
 */
char *s;
{
    static char *	result = (char *) 0;
    static char *	con    = (char *) 0;
    static char *	con_p  = (char *) 0;
    static int		n = 0;

    /*
     * Flush request ?!
     */
    if (s == (char *) 0) {
	result = (con && *con) ? con : "";
	if (con && *con) {
	    /*
	     * Force a new malloc
	     */
	    con   = (char *) 0;
	    con_p = (char *) 0;
	    n     = 0;
	}
	return result;
    }

    /*
     * Do we have enough space [3 = strlen(", ") + '\0']
     */
    if (con == (char *) 0 || (int)(con_p - con + strlen(s)) + 3 > n) {
	if (n == 0) {
	    con   = (char *) malloc(CONNECTCHUNK);
	    con_p = con;
	    *con = 0;
	} else {
	    int m = (int)(con_p - con);

	    con = (char *) realloc(con, n + CONNECTCHUNK);
	    con_p = con + m;
	}
	n += CONNECTCHUNK;
    }
    if (con == (char *) 0) {
	fprintf(stderr, "%s: Out of memory in con_cat()\n", program);
	exit(EXIT_FAILURE);
    }

    /*
     * Catenate the string
     */
    if (*con != 0) { strcat(con, ", "); con_p += 2; }
    strcat(con, s);			con_p += strlen(s);

    return (char *) 0;
}

/* #]con_cat:		*/ 
/* #[con_printf:	*/

#if defined (__STDC__)
	/*#[ANSI-C:	*/

#include <stdarg.h>

void
con_printf(char *fmt, ...)
{
    va_list argpoint;
    char msg[1024];

    va_start(argpoint, fmt);
	vsprintf(msg, fmt, argpoint);
    va_end(argpoint);
    con_cat(msg);
}

	/*#]ANSI-C:	*/ 
#else
	/*#[not ANSI-C:	*/

#include <varargs.h>

/*VARARGS1*/
void
con_printf(va_alist)
va_dcl
{
    va_list	argpoint;
    char *	fmt;
    char	msg[256];

    va_start(argpoint);
	fmt  = va_arg(argpoint, char *);
	vsprintf(msg, fmt, argpoint);
    va_end(argpoint);
    con_cat(msg);
}
	/*#]not ANSI-C:	*/ 
#endif

/* #]con_printf:	*/ 

/* #[board_cont:	*/

static void
board_cont(g)
/*
 * Update the way the board looks like based on the status
 */
GOGAME *g;
{
    PLACE p;

    for (p = 0; p < g->gm_sz*g->gm_sz; p++) {
	GOMOVE *s = g->gm_status  + p;
	GOMOVE *b = g->gm_board   + p;
	GOMOVE *i = g->gm_initial + p;

	/*
	 * Diagrams don't inherit letters, marks and numbers
	 */
	if (b->mv_letter) b->mv_letter = (CHAIN *) 0;
	if (b->mv_mark  ) b->mv_mark   = (CHAIN *) 0;
	b->mv_nr = 0;
	s->mv_nr = 0;

	/*
	 * Handle delayed moves, like captures, ko's, et cetera...
	 */
	b->mv_side = s->mv_side;
	i->mv_side = s->mv_side;
    }
}

/* #]board_cont:	*/ 
/* #[board_restore:	*/

static void
board_restore(g, m1, m2)
/*
 * Determine the board status based on the initial position and
 * the stones added/deleted between m1 and m2
 */
GOGAME *g;
GOMOVE *m1;
GOMOVE *m2;
{
    GOMOVE *m;
    CHAIN  *c;
    int     n = g->gm_sz * g->gm_sz;

    (void) memcpy(g->gm_status, g->gm_initial, n*sizeof(GOMOVE));
    (void) memcpy(g->gm_board , g->gm_initial, n*sizeof(GOMOVE));
    for (m = m1; m != m2; m = m->mv_next) {
/*	#[Addstone:		*/

	if (m->mv_addstone) {
	    for (c = m->mv_addstone; c; c = c->ch_next) {
		GOMOVE *added = c->ch_newstone;
		GOMOVE *s = g->gm_status + added->mv_place;

		s->mv_side = added->mv_side;
	    }
	}

/*	#]Addstone:		*/ 
/*	#[Delstone:		*/

	if (m->mv_delstone) {
	    for (c = m->mv_delstone; c; c = c->ch_next) {
		GOMOVE *emptied = c->ch_oldstone;
		GOMOVE *s = g->gm_status + emptied->mv_place;

		s->mv_side = 0;
	    }
	}

/*	#]Delstone:		*/ 
/*	#[Move:			*/

	if (IS_MOVE(m)) {
	    GOMOVE *s = g->gm_status + m->mv_place;

	    s->mv_side = m->mv_side;
	    s->mv_nr   = m->mv_nr;

	    /*
	     * The captured stones
	     */
	    for (c = m->mv_captured; c; c = c->ch_next) {
		GOMOVE *t = g->gm_status + c->ch_oldstone->mv_place;

		t->mv_side = 0;
	    }
	}
 
/*	#]Move:			*/ 
    }
}

/* #]board_restore:	*/ 
/* #[board_caption:	*/

char *
board_caption(g, m1, m2)
GOGAME *g;
GOMOVE *m1;
GOMOVE *m2;
{
    static char title[128];

    /*
     * Diagram has a name, we use it
     */
    if (g->gm_n && *(g->gm_n)) return strcpy(title, g->gm_n);

    /*
     * Diagram contains no moves
     */
    if (!m1) return strcpy(title, "No moves");

    /*
     * Diagram contains one move
     */
    if (m1 == m2) {
	sprintf(title, "%s %d",
		(m1->mv_side == BLACK) ? "B" : "W", m1->mv_nr);
	return title;
    }

    /*
     * Diagram contains two or more moves
     */
    sprintf(title, "%s %d -- %s %d",
	    (m1->mv_side == BLACK) ? "B" : "W", m1->mv_nr,
	    (m2->mv_side  == BLACK) ? "B" : "W", m2->mv_nr);
    return title;
}

/* #]board_caption:	*/ 
/* #[rnd2flat:		*/

static void
rnd2flat(g)
/*
 * Convert "round" sgf notation to "flattened" sgf notation:
 *
 *      |g                          |g
 *      |---|---|                   |---|
 *          |g  |v1                 |g  |v1
 *          |                       |
 *          |---|---|               |---|
 *              |g  |v1             |g  |v1
 *              |                   |
 *              |---|---|---|       |---|---|
 *                  |g  |v1 |v2     |g  |v1 |v2
 *                  |               |
 *
 *              round                flattened
 *              -----                ---------
 *
 * We add all first level variations to the game
 * We adjust the move numbers
 * We adjust the move total
 */
GOGAME *g;
{
    GOMOVE *m;
    int     movenr = 0;

    if (g == NULL) return;
    for (m = g->gm_move; m; m = m->mv_next) {
	GOGAME *h = m->mv_variation;
	GOGAME *i;
	GOMOVE *n;

	if (m->mv_nr) movenr = m->mv_nr;
	if (!h) continue;

	/*
	 * Increment the move numbers
	 */
	for (n = h->gm_move; n; n = n->mv_next) {
	    if (n->mv_nr) {
		n->mv_nr += movenr;
		(g->gm_movenr)++;
	    }
	}

	/*
	 * Link next diagram to main tree
	 */
	m->mv_variation = h->gm_next;

	/*
	 * Disable next diagram's ancestor
	 */
	if (h->gm_next) h->gm_next->gm_prev = (GOGAME *) 0;

	/*
	 * Adjust the parent for all variations on this level
	 */
	for (i = h->gm_next; i; i = i->gm_next) i->gm_parent = g;

	/*
	 * Add the moves to the main tree
	 */
	m->mv_next = h->gm_move;
    }
}

/* #]rnd2flat:		*/ 
/* #[seq2rnd:		*/

static void
seq2rnd(g)
/*
 * Convert "sequential" sgf notation to "round" sgf notation:
 *
 *      (g                              (g
 *       |--(---(                        |--(---(
 *          )v1 |g                          |g  )v1
 *              |                           |
 *              |---(---(                   |---(---(
 *                  )v1 |g                      |g  )v1
 *                      |                       |
 *                      |---(---(---(           |---(---(---(
 *                          )v1 )v2 |g              |g  )v1 )v2
 *                                  ))))            ))))
 *
 *              sequential                   round
 *              ----------                   -----
 *
 * We just insert the game in front of the variation list
 */
GOGAME *g;
{
    GOMOVE *m;

    if (g == NULL) return;
    for (m = g->gm_move; m; m = m->mv_next) {
	GOGAME *h;

	for (h = m->mv_variation; h; h = h->gm_next) {
	    if (h->gm_prev && h->gm_next == NULL) {
		/*
		 * We hit the game:
		 *	1. Move it in front
		 *	2. Treat it the same as its parent
		 */
		seq2rnd(h);
		h->gm_next      = m->mv_variation;
		m->mv_variation = h;
		if (h->gm_prev) {
		    h->gm_prev->gm_next = NULL;
		    h->gm_prev          = NULL;
		}
		break;
	    }
	}
    }
}

/* #]seq2rnd:		*/ 
/* #[game_print:	*/

/*
 * We insert automatic figure breaks when we encounter:
 *
 *	- Text
 *		in screen mode (no limit on the output)
 *		in kibitz mode
 *	- Letters
 *		i.e., if text is printed
 *	- Variations
 *
 * The break is set by setting "breakable" to 1. We do the actual
 * break when we encounter a node indicator to ensure that we
 * have collected all relevant information before exiting here.
 * However, when the user specified breaks (-f%50 for example)
 * we will never insert our breaks. We trust he knows what
 * he's doing.  Also in diagrams we never generate a break.
 */
#define FORCED_FIG	(f && figlist[*f])
#define IN_DIAGRAM	(!f)

static GOMOVE *
game_print(g, dev, m_next, f, d)
/*
 * Print a game starting at move m_first.
 * Return the move till where we printed.
 */
GOGAME	   *	g;
PRN_DEVICE *	dev;
GOMOVE	   *	m_next;		/* Next move to start printing	*/
int	   *	f;		/* requested figure  index	*/
int	   *	d;		/* requested diagram index	*/
{
    GOMOVE *	m;
    GOMOVE *	m_init = m_next;	/* The initial move		*/
    GOMOVE *	m_1st  = (GOMOVE *) 0;	/* 1st move with a number in it	*/
    GOMOVE *	m_lst  = (GOMOVE *) 0;	/* lst move with a number in it	*/
    GOMOVE *	m_dia  = (GOMOVE *) 0;	/* 1st move with a variation	*/
    GOMOVE *	m_txt  = (GOMOVE *) 0;	/* 1st move with a text block	*/
    GOMOVE *	board  = g->gm_board;	/* What's the initial position?	*/
    GOMOVE *	status = g->gm_status;	/* What's the initial status?	*/
    int		breakable = 0;
    unsigned   	moves = 0;   	    	/* Moves on the current figure  */

    /*
     * Allocate device resources
     */
    if (!(dev->dev_alloc)(g)) return (GOMOVE *) 0;

    /*
     * Set up the initial position
     */
    board_cont(g);

    /*
     * First, set up the whole position...
     */
    for (m = m_next; m; m_next = m->mv_next, m = m->mv_next) {
	CHAIN  *c;

/*	#[Node:			*/

	if (breakable) break;

/*	#]Node:			*/ 
/*	#[Addstone:		*/

	if (m->mv_addstone) {
	    for (c = m->mv_addstone; c; c = c->ch_next) {
		GOMOVE *added = c->ch_newstone;
		GOMOVE *b = board  + added->mv_place;
		GOMOVE *s = status + added->mv_place;

		b->mv_side = added->mv_side;
		s->mv_side = added->mv_side;
	    }
	}

/*	#]Addstone:		*/ 
/*	#[Delstone:		*/

	if (m->mv_delstone) {
	    for (c = m->mv_delstone; c; c = c->ch_next) {
		GOMOVE *emptied = c->ch_oldstone;
		GOMOVE *b = board  + emptied->mv_place;
		GOMOVE *s = status + emptied->mv_place;

		b->mv_side = 0;
		s->mv_side = 0;
	    }
	}

/*	#]Delstone:		*/ 
/*	#[Marks:		*/

	/*
	 * Humm, if we don't have text, what's the point of marks?!
	 */
	if (textout && m->mv_mark) {
	    /* breakable = 1; */
	    for (c = m->mv_mark; c; c = c->ch_next) {
		(board + c->ch_place)->mv_mark = c;
	    }
	}

/*	#]Marks:		*/ 
/*	#[Letters:		*/

	/*
	 * Humm, if we don't have text, what's the point of letters?!
	 */
	if (textout && m->mv_letter) {
	    if (!FORCED_FIG && !IN_DIAGRAM) breakable = 1;
	    for (c = m->mv_letter; c; c = c->ch_next) {
		(board + c->ch_place)->mv_letter = c;
	    }
	}

/*	#]Letters:		*/ 
/*	#[Move:			*/

	if (IS_MOVE(m) || IS_PASS(m)) {
	    /*
	     * Did we reach the required move number?
	     * If so, create a diagram
	     */
	    if (!IN_DIAGRAM) {
    	    	if (m->mv_nr == figlist[*f]) { (*f)++; breakable = 1; }
    	    	if (++moves == maxmoves) breakable = 1;
    	    }

	    /*
	     * The move markers
	     */
	    if (!m_1st) m_1st = m;
	    m_lst = m;

	    /*
	     * The move
	     */

    	    if (IS_PASS(m))
    	    	(dev->dev_pass)(g, m->mv_nr, m->mv_side);

    	    else {
       	    	GOMOVE *b = board  + m->mv_place;
    	    	GOMOVE *s = status + m->mv_place;

	    	s->mv_side = m->mv_side;
	    	s->mv_nr   = m->mv_nr;
	    	if (b->mv_side == FREE) {
		    b->mv_side = m->mv_side;
		    b->mv_nr   = m->mv_nr;
	    	}
	    	else {
		    if (b->mv_nr)	(dev->dev_onmove )
				    	(g, m->mv_nr, m->mv_side,
				            b->mv_nr, b->mv_side);
		    else		(dev->dev_onstone)
				    	(g, m->mv_nr,
					    m->mv_side,
					    m->mv_place);
	    	}

	    	/*
	     	 * The captured stones
	     	 */
	    	for (c = m->mv_captured; c; c = c->ch_next) {
		    GOMOVE *t = status + c->ch_oldstone->mv_place;

		    t->mv_side = 0;
	    	}
	    }
    	}
 
/*	#]Move:			*/ 
/*	#[Time:			*/

	if (m->mv_btime || m->mv_wtime)
		(dev->dev_time)(g, m->mv_btime, m->mv_wtime);

/*	#]Time:			*/ 
/*	#[Variations:		*/

	/*
	 * Remember the first moment with a variation
	 */
	if (m->mv_variation) {
	    if (!FORCED_FIG && !IN_DIAGRAM && *d == dialist[*d]) breakable = 1;
	    if (!m_dia) m_dia = m;
	}

/*	#]Variations:		*/ 
/*	#[Text:			*/

	/*
	 * Remember the first moment with text
	 * We only break when we do have any moves...
	 */
	if (textout && m->mv_text) {
	    /*
	     * The following test is explained in the function header
	     */
	    if (!FORCED_FIG		&&
		!IN_DIAGRAM		&&
		m_1st			&&	/* Any moves yet ?  */
		(kibitz || prn_dev == DEVICE_SCREEN)
	    ) breakable = 1; 
	    if (textout && !m_txt) m_txt = m;
	}

/*	#]Text:			*/ 
    }

    /*
     * Emission of the diagram
     */
    if (f || (*d == dialist[*d])) {
	if (!(dev->dev_emit) (g, m_1st, m_lst, m_txt, m_next)) {
	    (dev->dev_exit)();
	    exit(EXIT_SUCCESS);
	}
    }
    if (!f) (*d)++;

/*	#[Variations:	*/

    /*
     * Finally, show all variations referred to in the shown position...
     */
    if (m_dia) {
	for (m = m_dia; m != m_next; m = m->mv_next) {
	    if (m->mv_variation) {
		GOGAME *v;

		/*
		 * Restore actual board status from initial board status
		 */
		board_restore(g, m_init, m->mv_next);

		for (v = m->mv_variation; v; v = v->gm_next) {
		    GOMOVE *n = v->gm_move;

		    if (!board_alloc(v)) (dev->dev_exit)();
		    do { n = game_print(v, dev, n, NULL, d); } while (n);
		    board_free(v);
		}
	    }
	}
	/*
	 * Restore actual board status from initial board status
	 */
	board_restore(g, m_init, m_next);

    }

/*	#]Variations:	*/ 

    /*
     * Free the resources
     */
    (dev->dev_free)(g);

    /*
     * And return till where we ended
     */
    return m_next;
}

/* #]game_print:	*/ 
/* #[game_emit:		*/

static void
game_emit(g, dev)
GOGAME *g;
PRN_DEVICE *dev;
/*
 * Print/output a game on device dev
 */
{
    GOMOVE *  m = g->gm_move;
    int       f = 1;		/* First figure  from figlist[]	*/
    int       d = 1;		/* First diagram from dialist[]	*/

    if (!(dev->dev_init)()) (dev->dev_exit)();
    if (!board_alloc(g)) (dev->dev_exit)();

    do { m = game_print(g, dev, m, &f, &d); } while (m);

    board_free(g);
    (dev->dev_exit)();
}

/* #]game_emit:		*/ 

