/* #[info:		*/
/************************************************************************
*									*
*			 ,		,,				*
*			||		||				*
*		       =||=  /'\\  /'\\ ||  _-_,			*
*			||  || || || || || ||_.				*
*			||  || || || || ||  ~ ||			*
*			\\, \\,/  \\,/	\\ ,-_-				*
*									*
*									*
*									*
*		  Copyright (c) 1993  Jan van der Steen			*
*	       Centre for Mathematics and Computer Science		*
*		       Amsterdam, the Netherlands			*
*									*
*-----------------------------------------------------------------------*
*   You are not allowed to use this software in commercial software.	*
*   You are allowed to use this software in free software under the	*
*   condition that you maintain this integral message in the source.	*
*-----------------------------------------------------------------------*
* File	  : tools.c							*
* Purpose : Various programming tools					*
* Version : 1.9								*
* Modified: 9/22/93 17:20:50						*
* Author  : Jan van der Steen (jansteen@cwi.nl)				*
************************************************************************/
/* #]info:		*/ 
/* #[include:		*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tools.h"

/* #]include:		*/ 
/* #[prototype:		*/

#ifdef __STDC__
#       define  PROTO(s) s
#else
#       define  PROTO(s) ()
#endif

static void	message_emit	PROTO((char *s));

#undef PROTO

/* #]prototype:		*/ 

/* #[string_alloc:	*/

char *
string_alloc(s)
char *s;
{
    char *t = (char *) malloc(strlen(s)+1);

    if (t == (char *) 0) {
        message("Out of memory in string_alloc\n");
        exit(1);
    }
    return t;
}

/* #]string_alloc:	*/ 
/* #[string_store:	*/

char *
string_store(s)
char *s;
{
    return strcpy(string_alloc(s), s);
}

/* #]string_store:	*/ 
/* #[message_emit:	*/

static void
message_emit(s)
char *s;
{
    static long emitted = 0;

    if (emitted + strlen(s) > 72) {
	fputc('\n', stderr);
	emitted = 0;
    }
    fprintf(stderr, "%s", s);
    emitted += strlen(s);
}

/* #]message_emit:	*/ 
/* #[message:		*/

#if defined (__STDC__)
	/*#[ANSI-C:	*/
#include <stdarg.h>

void
message(char *fmt, ...)
{
    va_list argpoint;
    char msg[2024];

    va_start(argpoint, fmt);
	vsprintf(msg, fmt, argpoint);
    va_end(argpoint);
    message_emit(msg);
}

	/*#]ANSI-C:	*/ 
#else
	/*#[not ANSI-C:	*/

#include <varargs.h>

/*VARARGS1*/
void
message(va_alist)
va_dcl
{
    va_list	argpoint;
    char *	fmt;
    char	msg[2024];

    va_start(argpoint);
	fmt  = va_arg(argpoint, char *);
	vsprintf(msg, fmt, argpoint);
    va_end(argpoint);
    message_emit(msg);
}
	/*#]not ANSI-C:	*/ 
#endif

/* #]message:		*/ 

/* #[str2rank:		*/

#define RANK_ILLEGAL	1000000L	/* Make them noticable */
#define RANK_PRO	 100000L
#define RANK_IGS	  10000L
#define RANK_DAN	   1000L
#define RANK_KYU	     10L
#define RANK_NR		      2L
#define RANK_UNKNOWN	      1L

long
str2rank(s)
/*
 * Parse and calculate a rank.
 * We distinguish seven categories, positive if ranked, negative if rated:
 *
 *	Sample		Our spec	Explanation
 *	-------------------------------------------
 *	<any>		RANK_ILLEGAL	Not defined
 *	"3p "		RANK_PRO	Professionals
 *	"IGS??"		RANK_IGS	IGS champion
 *	"4d*"		RANK_DAN	dan players
 *	"15k "		RANK_KYU	kyu players
 *	"NR "		RANK_NR		unranked players
 *	"???"		RANK_UNKNOWN	IGS unknowns
 */
char *s;
{
    char *t = s;
    long  factor = 0;
    int   rank   = 0;
    int   rated  = 1;

    if (!s || !*s) return RANK_ILLEGAL;
    s += strlen(s)-1;
    if (*s == '*') rated = -1;
    if (*s == ' ' || *s == '*') s--;
    switch (*s) {
	case 'p': factor = RANK_PRO; break;
	case 'd': factor = RANK_DAN; break;
	case 'k': factor = RANK_KYU; break;
	case 'R': return rated * RANK_NR;
	case '?': return rated * RANK_UNKNOWN;
	default :
	    if ('1' <= *s || *s <= '9')
		return rated* (*s) * RANK_IGS;	/* Assume IGS champion */
	    else
		return rated*RANK_ILLEGAL;	/* Assume illegal rating */
    }

    /*
     * When we arrive here we have to parse a {p,d,k} rank
     */
    while ('0' <= *t && *t <= '9') rank = 10*rank + (*t++ - '0');

    return rated * factor * rank;
}

/* #]str2rank:		*/ 
/* #[rank2str:		*/

char *
rank2str(rank)
long rank;
{
    static char s[10];
    int		rated = 0;

    if (rank < 0) { rated = 1; rank *= -1; }

    if (rank == RANK_ILLEGAL)	strcpy (s, "XX"); else
    if (rank == RANK_UNKNOWN)	strcpy (s, "??"); else
    if (rank == RANK_NR     )	strcpy (s, "NR"); else
    if (rank >= RANK_IGS    )	sprintf(s, "IGS9%ld", rank/RANK_IGS); else
    if (rank >= RANK_PRO    )	sprintf(s, "%2ldp"  , rank/RANK_PRO); else
    if (rank >= RANK_DAN    )	sprintf(s, "%2ldd"  , rank/RANK_DAN); else
    if (rank >= RANK_KYU    )	sprintf(s, "%2ldk"  , rank/RANK_KYU);
    if (rated) strcat(s, "*");

    return s;
}

/* #]rank2str:		*/ 

